﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public interface ISwichClient
{
    void OnSwitchStateChanged(bool newState);
}

public class Switch : FocusHolder
{
    [System.Serializable]
    public class SwitchState
    {
        public GameObject effectObj;
        public Color color;
    }

    public List<GameObject> linkedObjects;
    public float stareThreshold;

    public GameObject nodePrefab;
    public float spawnNodeInterval;

    public SwitchState stateOn;
    public SwitchState stateOff;
    public List<SpriteRenderer> colorObjects;

    public float pathOffsetForSelf;
    public LineRenderer lineRendererPrefab;

    private bool currentState;
    private float spawnNodeIn;

    private List<List<Transform>> switchPaths;
    private List<LineRenderer> lineRenderers;

    void Start()
    {
        switchPaths = new List<List<Transform>>();
        lineRenderers = new List<LineRenderer>();

        foreach (var linkedObject in linkedObjects)
        {
            List<Transform> switchPath = new List<Transform>();
            switchPath.Add(transform);
            switchPath.Add(linkedObject.transform);
            switchPaths.Add(switchPath);
        }

        foreach (var switchPath in switchPaths)
        {
            var lineRenderer = Instantiate(lineRendererPrefab.gameObject).GetComponent<LineRenderer>();

            List<Vector3> positions = new List<Vector3>();
            for (int index = 0; index < switchPath.Count; index++)
            {
                var trans = switchPath[index];
                Vector3 pos = trans.position;
                if (trans == transform && index < switchPath.Count - 1)
                {
                    var dir = switchPath[index + 1].position - pos;
                    dir.Normalize();
                    pos += dir * pathOffsetForSelf;
                }

                positions.Add(pos);
            }

            lineRenderer.numPositions = positions.Count;
            lineRenderer.SetPositions(positions.ToArray());

            lineRenderers.Add(lineRenderer);
        }
        
        ChangeState(false);
    }

    private void SpawnNode()
    {
        foreach (var switchPath in switchPaths)
        {
            var node = Instantiate(nodePrefab, transform.position, Quaternion.identity).GetComponent<SwitchNode>();
            node.spRenderer.color = GetActiveState().color;
            node.path = switchPath;
        }

        spawnNodeIn = spawnNodeInterval;
    }

    protected override void Update()
    {
        base.Update();

        if (currentState)
        {
            spawnNodeIn -= Time.deltaTime;
            if (spawnNodeIn <= 0)
            {
                SpawnNode();
            }
        }

        if (HasFocus != currentState)
        {
            ChangeState(HasFocus);
        }
    }

    private SwitchState GetActiveState()
    {
        return currentState? stateOn : stateOff;
    }

    private void ChangeState(bool newState)
    {
        foreach (var linkedObject in linkedObjects)
        {
            var clent = linkedObject.GetComponent<ISwichClient>();
            if (clent != null)
            {
                clent.OnSwitchStateChanged(newState);
            }
        }

        currentState = newState;

        var activeState = GetActiveState();
        var inactiveState = currentState ? stateOff : stateOn;

        inactiveState.effectObj.SetActive(false);
        activeState.effectObj.SetActive(true);

        foreach (var colorObject in colorObjects)
        {
            colorObject.color = activeState.color;
        }

        foreach (var lineRenderer in lineRenderers)
        {
            lineRenderer.startColor = activeState.color;
            lineRenderer.endColor = activeState.color;
        }

        ResetFocusTimer();
    }
}
